// KeypadManager.cpp : Implementation of CKeypadManager
#include "stdafx.h"
#include "ADSmartIO.h"
#include "KeypadManager.h"
#include "SmartIOHelper.h"
/////////////////////////////////////////////////////////////////////////////
// CKeypadManager

UINT			g_KeyMap[MAXKEYS - 1];
UINT			g_Repeat;
unsigned long	threadID;
VARIANT_BOOL	bDebug;

CKeypadManager::CKeypadManager()
{
	// link to SmartIO
	if(!LoadSmartIO())
	{
		this->Error(_T("Failed to load SmartIO library!!"));
	}
	else
	{
		// disable debugging
		bDebug = VARIANT_FALSE;

		// turn on keypad
		SIOSelectOption(1);

		// set the number of keys
		m_rows = 8;
		m_cols = 8;

		if(SIOSetKeyPadSize(m_cols, m_rows))
		{
			RETAILMSG(TRUE,(_T("Setting defaults...\r\n")));

			// initialize global array
			this->ClearKeypadMap();

			// set defaults
			g_Repeat = 100;
			lDelay = 40;
			threadID = 0;
		}
		
		if(bDebug)
			RETAILMSG(TRUE,(_T("\r\nKeypad Manager Loaded successfully %d.\r\n"), SIOFirmwareVersion()));
	}
}

CKeypadManager::~CKeypadManager()
{
	if(hScanTimer)
	{
		if(bDebug)
			RETAILMSG(TRUE, (_T("killing timer: \r\n")));	

		KillTimer(NULL, hScanTimer);
	}
	
	if(threadID)
	{
		if(bDebug)
			RETAILMSG(TRUE, (_T("terminating thread: \r\n")));	

		TerminateThread(&threadID, 0);
	}

	if(bDebug)
		RETAILMSG(TRUE,(L"Keypad Manager unloaded\r\n"));
}

STDMETHODIMP CKeypadManager::EnableKeypad()
{
	ThreadParam t;;
	t.delay = lDelay;

	if(bDebug)
		RETAILMSG(TRUE, (_T("creating thread: \r\n")));	

	CreateThread(NULL, NULL, &CKeypadManager::ThreadMarshal, (LPVOID)this, 0, &threadID);
	
	if(bDebug)
		RETAILMSG(TRUE,(L"Keypad Manager enabled.\r\n"));

	return S_OK;
}

STDMETHODIMP CKeypadManager::DisableKeypad()
{
	if(hScanTimer)
	{
		if(bDebug)
			RETAILMSG(TRUE, (_T("killing timer: \r\n")));	

		KillTimer(NULL, hScanTimer);
	}

	if(threadID)
	{
		if(bDebug)
			RETAILMSG(TRUE, (_T("posting message to kill thread: \r\n")));	

		PostThreadMessage(threadID, WM_QUIT, 0, 0);
	}

	if(bDebug)
		RETAILMSG(TRUE,(L"Keypad Manager disabled.\r\n"));

	return S_OK;
}

STDMETHODIMP CKeypadManager::MapKeypad(int KeyNumber, int VirtualKey)
{
	g_KeyMap[KeyNumber] = VirtualKey;

	if(bDebug)
		RETAILMSG(TRUE,(_T("Keypad Manager mapped key %d to %d.\r\n"), KeyNumber, VirtualKey));

	return S_OK;
}

STDMETHODIMP CKeypadManager::ClearKeypadMap()
{
	// clear global array
	for(int i = 0 ; i < MAXKEYS - 1 ; i++)
		g_KeyMap[i] = 0;

	if(bDebug)
		RETAILMSG(TRUE,(L"Keypad Manager map cleared.\r\n"));

	return S_OK;
}

STDMETHODIMP CKeypadManager::get_KeypadDelay(long *pVal)
{
	if(SIOFirmwareVersion() == 16407)
		return Error(_T("Delay not supported on Firmware 0x4017"));

	*pVal = lDelay;

	return S_OK;
}

STDMETHODIMP CKeypadManager::put_KeypadDelay(long newVal)
{
	if(SIOFirmwareVersion() == 16407)
		return Error(_T("Delay not supported on Firmware 0x4017"));

	if((newVal > 10000) || (newVal < 1))
		return Error(_T("Repeat value out of range. (must be 1-10000)"));
	
	lDelay = newVal;

	if(bDebug)
		RETAILMSG(TRUE,(_T("Keypad Manager delay changed to %d.\r\n"), newVal));

	// check to see if a timer is running
	if(threadID)
	{
		__try
		{
			if(hScanTimer)
			{
				if(bDebug)
					RETAILMSG(TRUE, (_T("killing timer: \r\n")));	
			
				KillTimer(NULL, hScanTimer);
			}

			if(bDebug)
				RETAILMSG(TRUE, (_T("posting message to kill thread: \r\n")));	

			// kill existing thread
			PostThreadMessage(threadID, WM_QUIT, 0, 0);

			if(WaitForSingleObject(&threadID, 5000) == WAIT_TIMEOUT)
				RETAILMSG(TRUE,(_T("Keypad Manager timeout elapsed waiting for thread termination.\r\n")));

			// start a new thread
			ThreadParam t;;
			t.delay = lDelay;
			
			if(bDebug)
				RETAILMSG(TRUE, (_T("creating new thread \r\n")));	

			CreateThread(NULL, NULL, &CKeypadManager::ThreadMarshal, (LPVOID)this, 0, &threadID);
		}
		__except(EXCEPTION_EXECUTE_HANDLER)
		{
			RETAILMSG(TRUE, (_T("Exception restarting thread: %d"), _exception_code()));
		}
	}

	return S_OK;
}

STDMETHODIMP CKeypadManager::get_KeypadRepeat(long *pVal)
{
	if(SIOFirmwareVersion() == 16407)
		return Error(_T("Repeat not supported on Firmware 0x4017"));

	*pVal = g_Repeat;

	return S_OK;
}

STDMETHODIMP CKeypadManager::put_KeypadRepeat(long newVal)
{
	if(SIOFirmwareVersion() == 16407)
		return Error(_T("Repeat not supported on Firmware 0x4017"));

	if((newVal > 5000) || (newVal < 1))
		return Error(_T("Repeat value out of range. (must be 1-5000)"));

	g_Repeat = newVal;

	if(bDebug)
		RETAILMSG(TRUE,(_T("Keypad Manager repeat changed to %d.\r\n"), newVal));

	return S_OK;
}

void CKeypadManager::ScanKeypad()
{
	MSG msg;

	// set the timer
	hScanTimer = SetTimer(NULL, 0, lDelay, &CKeypadManager::KeypadTimer);

	// message pump
	while (GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);

	}
}

DWORD WINAPI CKeypadManager::ThreadMarshal(LPVOID pParam)
{
	CKeypadManager *pThis = (CKeypadManager *)pParam;
	
	if(bDebug)
		RETAILMSG(TRUE, (_T("thread marshalled: \r\n")));

	pThis->ScanKeypad();

	if(bDebug)
		RETAILMSG(TRUE, (_T("thread ending: \r\n")));	

	// let app know we've exited
	threadID = 0;

	return 0;
}

void CALLBACK CKeypadManager::KeypadTimer(HWND hwnd, UINT uMsg, UINT idEvent, DWORD dwTime)
{	
	static int lastkey;
	static DWORD tick;

	// scan keyboard
	BYTE keydata = SIOReadKeypad();

	// return on timeout
	if(keydata == 0xFF)
	{
		lastkey = 0;
		tick = 0;
		
		if(bDebug)
			RETAILMSG(TRUE, (_T("sioreadkeypad timeout: \r\n")));

		return;
	}
	if(bDebug)
		RETAILMSG(TRUE, (_T("Key: %d\r\n"), keydata));
	
	if(keydata > 63)
		return;

	if((lastkey == keydata) && ((GetTickCount() - tick) < g_Repeat))
	{
		if(bDebug)
			RETAILMSG(TRUE, (_T("same key or under repeat time\r\n")));

		return;
	}
	else
	{
		if(bDebug)
			RETAILMSG(TRUE, (_T("updating tick count and last key: \r\n")));

		tick = GetTickCount();
		lastkey = keydata;
	}
		
	if(g_KeyMap[keydata] != 0)
	{
		if(bDebug)
			RETAILMSG(TRUE, (_T("sending key: %d\r\n"), g_KeyMap[keydata]));

		RETAILMSG(TRUE,(L"g_KeyMap[%i] = %i\r\n", keydata, g_KeyMap[keydata]));

		keybd_event(g_KeyMap[keydata], 0 ,0, 0);
	}

	return;
}


BOOL LoadSmartIO()
{
	RETAILMSG(TRUE,(L"\r\nLoading SMARTIO.DLL...\r\n"));
	gLibSMARTIO = LoadLibrary((L"SMARTIO.dll"));

	if(gLibSMARTIO == NULL)
	{	
		RETAILMSG(TRUE,(L"\r\nCan't find SMARTIO.DLL\r\n"));
		RETAILMSG(TRUE,(L"\r\nLoading KEYBDDR.DLL...\r\n"));

		gLibKEYBDDR = LoadLibrary((L"KEYBDDR.dll"));

		if(gLibKEYBDDR == NULL)
		{	
			RETAILMSG(TRUE,(L"\r\nFailed to load ADSmartIO library!\r\n"));
			return FALSE;
		}
		else
		{	
			RETAILMSG(TRUE,(L"\r\nKEYBDDR.DLL Loaded successfully...\r\n"));
			gLibSMARTIO = gLibKEYBDDR;
		}
	}
	else
	{	
		RETAILMSG(TRUE,(L"\r\nSMARTIO.DLL Loaded successfully...\r\n"));
	}
	if(gLibSMARTIO!=NULL)
	{	
		SIOFirmwareVersion = (SIOFIRMWAREVERSION)GetProcAddress(gLibSMARTIO,(L"SIOFirmwareVersion"));
		SIOReadKeypad = (SIOREADKEYPAD)GetProcAddress(gLibSMARTIO,(L"SIOReadKeypad"));
		SIOSetKeyPadSize = (SIOSETKEYPADSIZE)GetProcAddress(gLibSMARTIO,(L"SIOSetKeyPadSize"));
		SIOSelectOption = (SIOSELECTOPTION)GetProcAddress(gLibSMARTIO,(L"SIOSelectOption"));
/*
		SIOInitDDRA = (SIOINITDDRA)GetProcAddress(gLibSMARTIO,(L"SIOInitDDRA"));
		SIOInitDDRB = (SIOINITDDRB)GetProcAddress(gLibSMARTIO,(L"SIOInitDDRB"));
		SIOInitDDRC = (SIOINITDDRC)GetProcAddress(gLibSMARTIO,(L"SIOInitDDRC"));
		SIOInitDDRD = (SIOINITDDRD)GetProcAddress(gLibSMARTIO,(L"SIOInitDDRD"));
		SIOReadPortA = (SIOREADPORTA)GetProcAddress(gLibSMARTIO,(L"SIOReadPortA"));
		SIOReadPortB = (SIOREADPORTB)GetProcAddress(gLibSMARTIO,(L"SIOReadPortB"));
		SIOReadPortC = (SIOREADPORTC)GetProcAddress(gLibSMARTIO,(L"SIOReadPortC"));
		SIOReadPortD = (SIOREADPORTD)GetProcAddress(gLibSMARTIO,(L"SIOReadPortD"));
		SIOWritePortA = (SIOWRITEPORTA)GetProcAddress(gLibSMARTIO,(L"SIOWritePortA"));
		SIOWritePortB = (SIOWRITEPORTB)GetProcAddress(gLibSMARTIO,(L"SIOWritePortB"));
		SIOWritePortC = (SIOWRITEPORTC)GetProcAddress(gLibSMARTIO,(L"SIOWritePortC"));
		SIOWritePortD = (SIOWRITEPORTD)GetProcAddress(gLibSMARTIO,(L"SIOWritePortD"));
*/
	}

	return TRUE;

}

STDMETHODIMP CKeypadManager::get_KeypadRows(int *pVal)
{
	*pVal = m_rows;

	return S_OK;
}

STDMETHODIMP CKeypadManager::put_KeypadRows(int newVal)
{
	if((newVal > 8) || (newVal < 1))
		return Error(_T("Row value out of range. (must be 1-8)"));
	
	if(SIOSetKeyPadSize(m_cols, newVal))
	{
		m_rows = newVal;

		if(bDebug)
			RETAILMSG(TRUE,(_T("Keypad Manager rows changed to %d.\r\n"), newVal));
	}

	return S_OK;
}

STDMETHODIMP CKeypadManager::get_KeypadColumns(int *pVal)
{
	*pVal = m_cols;

	return S_OK;
}

STDMETHODIMP CKeypadManager::put_KeypadColumns(int newVal)
{
	if((newVal > 8) || (newVal < 1))
		return Error(_T("Column value out of range. (must be 1-8)"));
	
	if(SIOSetKeyPadSize(newVal, m_rows))
	{
		m_cols = newVal;
		
		if(bDebug)
			RETAILMSG(TRUE,(_T("Keypad Manager columns changed to %d.\r\n"), newVal));
	}

	return S_OK;
}

STDMETHODIMP CKeypadManager::get_ShowDebugMessages(VARIANT_BOOL *pVal)
{
	*pVal = bDebug;

	return S_OK;
}

STDMETHODIMP CKeypadManager::put_ShowDebugMessages(VARIANT_BOOL newVal)
{
	bDebug = newVal;

	return S_OK;
}
